@@ -11,6 +11,7 @@ import shutil
 import tempfile
 import typing
 import zipfile
+import zlib
 from collections.abc import Callable, Generator
 from pathlib import Path
 from typing import Any, Literal
@@ -19,6 +20,7 @@ import boto3
 import fsspec
 import fsspec.utils
 from cloudpathlib import CloudPath, S3Client, S3Path
+from iker.common.utils.sequtils import chunk_between, head, last
 from iker.common.utils.shutils import glob_match, listfile, path_depth
 from iker.common.utils.strutils import is_empty, trim_to_none
 from rich.progress import BarColumn, DownloadColumn, Progress, TaskID, TextColumn, TransferSpeedColumn
@@ -37,9 +39,10 @@ __all__ = [
     "S3TransferCallbackClient",
     "s3_make_progress_callback",
     "s3_make_progressed_client",
-    "s3_estimate_archive_ranged_requests",
-    "s3_open_archive_member",
-    "s3_open_archive_members",
+    "s3_archive_member_tree",
+    "s3_archive_listfile",
+    "s3_archive_open_member",
+    "s3_archive_open_members",
 ]
 
 
@@ -86,7 +89,12 @@ def s3_make_client(
             client.close()
 
 
-def s3_list_objects(client: S3Client, bucket: str, prefix: str, limit: int = None) -> Generator[S3ObjectMeta]:
+def s3_list_objects(
+    client: S3Client,
+    bucket: str,
+    prefix: str,
+    limit: int = None,
+) -> Generator[S3ObjectMeta, None, None]:
     """
     Lists all objects from the given S3 ``bucket`` and ``prefix``.
 
@@ -128,7 +136,7 @@ def s3_listfile(
     include_patterns: list[str] | None = None,
     exclude_patterns: list[str] | None = None,
     depth: int = 0,
-) -> Generator[S3ObjectMeta]:
+) -> Generator[S3ObjectMeta, None, None]:
     """
     Lists all objects from the given S3 ``bucket`` and ``prefix``, filtered by patterns and directory depth.
 
@@ -424,7 +432,7 @@ def s3_make_progressed_client(
     secret_access_key: str = None,
     region_name: str = None,
     endpoint_url: str = None,
-) -> Generator[S3Client]:
+) -> Generator[S3Client, None, None]:
     """
     Creates an S3 client with progress callback as a context manager for safe resource handling.
 
@@ -481,70 +489,152 @@ def s3_options_from_s3_client(client: S3Client) -> dict[str, Any]:
     return s3_options
 
 
-def s3_estimate_archive_ranged_requests(
+type ArchiveMemberTree = dict[str, tuple[zipfile.ZipInfo, ArchiveMemberTree | None]]
+
+
+def s3_archive_member_tree(
     client: S3Client,
-    archive_url: str,
-    members: list[str],
-    *,
-    central_directory_overhead: int = 64 * 1024,
-    member_header_overhead: int = 128,
-) -> tuple[int, int, list[str]]:
+    bucket: str,
+    key: str,
+) -> ArchiveMemberTree:
     """
-    Estimates the total size of ranged requests needed to extract specific members from a ZIP archive stored in S3.
+    Builds a tree structure of members in a ZIP archive stored in S3 for efficient lookup.
+    Directories have ZipInfo and a nested dict; files have ZipInfo and None.
+    Directory members are recognized by names ending with a trailing slash ("/").
+
+    Example:
+    {
+      "dir1/": (ZipInfo, {
+          "file1.txt": (ZipInfo, None),
+          "subdir/": (ZipInfo, {
+              "file2.txt": (ZipInfo, None)
+          })
+      }),
+      "file3.txt": (ZipInfo, None)
+    }
 
     :param client: An instance of ``S3Client``.
-    :param archive_url: The URL of the ZIP archive in S3.
-    :param members: List of member file names to extract from the archive.
-    :param central_directory_overhead: Estimated overhead size for the central directory.
-    :param member_header_overhead: Estimated overhead size for each member's header.
-
-    :return: A tuple containing:
-             - Estimated total size of ranged requests (in bytes).
-             - Total size of the archive (in bytes).
-             - List of member names that were not found in the archive.
+    :param bucket: Bucket name.
+    :param key: Object key of the ZIP archive.
+    :return: A tree structure of members in the ZIP archive.
     """
     s3_options = s3_options_from_s3_client(client)
 
-    storage_opts = fsspec.utils.infer_storage_options(archive_url)
-    protocol = storage_opts.get("protocol")
-    if protocol != "s3":
-        raise ValueError(f"unsupported protocol '{protocol}', only 's3' is supported")
+    archive_url = f"s3://{bucket}/{key}"
+
+    with fsspec.open(archive_url, "rb", s3=s3_options) as s3_fh, zipfile.ZipFile(s3_fh) as archive:
+        member_zip_infos = archive.infolist()
+
+    root_member_tree: ArchiveMemberTree = {}
+
+    def build_member_tree(info: zipfile.ZipInfo):
+        *parts, last_part = info.filename.rstrip("/").split("/")
+        current_member_tree = root_member_tree
+        for part in parts:
+            _, current_member_tree = current_member_tree.setdefault(part + "/", (None, {}))
+        if info.is_dir():
+            current_member_tree[last_part + "/"] = info, {}
+        else:
+            current_member_tree[last_part] = info, None
+
+    # Sort by filename to ensure directories are created before their contents
+    for info in sorted(member_zip_infos, key=lambda x: x.filename):
+        build_member_tree(info)
+
+    return root_member_tree
+
+
+def s3_archive_listfile(
+    client: S3Client,
+    bucket: str,
+    key: str,
+    members: list[str] | None = None,
+) -> tuple[int, list[zipfile.ZipInfo], list[str]]:
+    """
+    Lists members of a ZIP archive stored in S3, optionally filtering by specific member names. When filtering,
+    if a member is a directory, it must end with a trailing slash ("/") to be recognized as such, and all files
+    under that directory will be included in the results.
+
+    Example usage:
+    >>> archive_size, member_zip_infos, missed_members = s3_archive_listfile(client, bucket, key, members=["file1.txt", "dir1/"])
+    >>> for info in member_zip_infos:
+    ...     print(info.filename, info.file_size)
+    >>> if missed_members:
+    ...     print("Members not found:", missed_members)
+
+    :param client: An instance of ``S3Client``.
+    :param bucket: Bucket name.
+    :param key: Object key of the ZIP archive.
+    :param members: Optional list of member names to filter; if ``None``, all members are returned.
+    :return: A tuple containing:
+             - The size of the archive in bytes.
+             - A list of ``zipfile.ZipInfo`` objects for the included members.
+             - A list of member names that were not found in the archive.
+    """
+    s3_options = s3_options_from_s3_client(client)
+
+    archive_url = f"s3://{bucket}/{key}"
 
     fs = fsspec.filesystem("s3", **s3_options)
-
     archive_size = fs.size(archive_url)
 
-    ranged_total_size = 0
-    missing_members = []
+    root_member_tree = s3_archive_member_tree(client, bucket, key)
 
-    with fsspec.open(archive_url, "rb", s3=s3_options) as fh, zipfile.ZipFile(fh) as archive:
-        # central directory read happened here (small)
-        for member in members:
-            try:
-                member_info = archive.getinfo(member)
-                ranged_total_size += member_info.compress_size + member_header_overhead
-            except KeyError:
-                missing_members.append(member)
+    def search_members_tree(member: str) -> tuple[zipfile.ZipInfo | None, dict | None]:
+        *parts, last_part = member.rstrip("/").split("/")
+        current_member_tree = root_member_tree
+        for part in parts:
+            _, current_member_tree = current_member_tree.get(part + "/", (None, None))
+            if current_member_tree is None:
+                return None, None
+        if member.endswith("/"):  # Directory member recognized by trailing slash
+            return current_member_tree.get(last_part + "/", (None, None))
+        else:
+            return current_member_tree.get(last_part, (None, None))
+
+    def collect_member_zip_infos(tree: ArchiveMemberTree) -> Generator[zipfile.ZipInfo, None, None]:
+        for member_zip_info, member_tree in tree.values():
+            if member_zip_info is None:
                 continue
+            if member_zip_info.is_dir():
+                yield from collect_member_zip_infos(member_tree)
+            else:
+                yield member_zip_info
 
-    ranged_total_size += central_directory_overhead
+    if members is None:
+        return archive_size, list(collect_member_zip_infos(root_member_tree)), []
 
-    return ranged_total_size, archive_size, missing_members
+    included_member_zip_infos = []
+    missed_members = []
+
+    for member in members:
+        member_zip_info, member_tree = search_members_tree(member)
+        if member_zip_info is None:
+            missed_members.append(member)
+            continue
+        if not member_zip_info.is_dir():
+            included_member_zip_infos.append(member_zip_info)
+        else:
+            included_member_zip_infos.extend(collect_member_zip_infos(member_tree or {}))
+
+    return archive_size, included_member_zip_infos, missed_members
 
 
 if typing.TYPE_CHECKING:
-    def s3_open_archive_member(
+    def s3_archive_open_member(
         client: S3Client,
-        archive_url: str,
+        bucket: str,
+        key: str,
         member: str,
         mode: Literal["r", "rb"] = "r",
     ) -> contextlib.AbstractContextManager[typing.IO]: ...
 
 
 @contextlib.contextmanager
-def s3_open_archive_member(
+def s3_archive_open_member(
     client: S3Client,
-    archive_url: str,
+    bucket: str,
+    key: str,
     member: str,
     mode: Literal["r", "rb"] = "r",
 ) -> Generator[typing.IO, None, None]:
@@ -552,32 +642,46 @@ def s3_open_archive_member(
     Opens a specific member file from a ZIP archive stored in S3.
 
     :param client: An instance of ``S3Client``.
-    :param archive_url: The URL of the ZIP archive in S3.
+    :param bucket: Bucket name.
+    :param key: Object key of the ZIP archive.
     :param member: The member file name to open from the archive.
     :param mode: File mode for opening the member ("r" for text, "rb" for binary).
 
     :return: A file-like object for the specified member within the ZIP archive.
     """
+    if mode not in ("r", "rb"):
+        raise ValueError("mode must be either 'r' or 'rb'")
+
     s3_options = s3_options_from_s3_client(client)
 
-    storage_opts = fsspec.utils.infer_storage_options(archive_url)
-    protocol = storage_opts.get("protocol")
-    if protocol != "s3":
-        raise ValueError(f"unsupported protocol '{protocol}', only 's3' is supported")
-
-    with fsspec.open(f"zip://{member}::{archive_url}", mode, s3=s3_options) as fh:
-        yield fh
+    with fsspec.open(f"zip://{member}::s3://{bucket}/{key}", mode, s3=s3_options) as s3_fh:
+        yield s3_fh
 
 
-def s3_open_archive_members(
+ZIPFILE_HEADER_MIN_SIZE = 30
+ZIPFILE_HEADER_FN_LEN_OFFSET = 26
+ZIPFILE_HEADER_EX_LEN_OFFSET = 28
+
+
+def zipfile_hdr_fn_len_slice(offset: int) -> slice:
+    return slice(offset + ZIPFILE_HEADER_FN_LEN_OFFSET, offset + ZIPFILE_HEADER_FN_LEN_OFFSET + 2)
+
+
+def zipfile_hdr_ex_len_slice(offset: int) -> slice:
+    return slice(offset + ZIPFILE_HEADER_EX_LEN_OFFSET, offset + ZIPFILE_HEADER_EX_LEN_OFFSET + 2)
+
+
+def s3_archive_open_members(
     client: S3Client,
-    archive_url: str,
-    members: list[str],
+    bucket: str,
+    key: str,
+    members: list[str] | None = None,
     mode: Literal["r", "rb"] = "r",
     *,
     threshold: float = 0.5,
     central_directory_overhead: int = 64 * 1024,
     member_header_overhead: int = 128,
+    use_chunked_reads: bool = False,
 ) -> Generator[tuple[str, Callable[[], typing.IO]], None, None]:
     """
     Choose the best transfer strategy (ranged requests per-member vs full archive transfer)
@@ -590,67 +694,138 @@ def s3_open_archive_members(
 
     Example usage:
 
-    >>> for member, opener in s3_open_archive_members(client, archive_url, members):
+    >>> for member, opener in s3_archive_open_members(client, bucket, key, members):
     ...     with opener() as fh:
     ...         data = fh.read()
 
     Incorrect usage that may lead to errors due to temporary file cleanup:
 
     >>> openers = []
-    >>> for member, opener in s3_open_archive_members(client, archive_url, members):
+    >>> for member, opener in s3_archive_open_members(client, bucket, key, members):
     ...     openers.append((member, opener))
     >>> for member, opener in openers:
     ...     with opener() as fh:  # May fail if temporary files have been cleaned up
     ...         data = fh.read()
 
     :param client: An instance of ``S3Client``.
-    :param archive_url: S3 URL to the ZIP archive.
+    :param bucket: Bucket name.
+    :param key: Object key of the ZIP archive.
     :param members: List of member names to stream.
     :param mode: File mode for opening members ("r" for text, "rb" for binary).
     :param threshold: If (estimated ranged transfer bytes / archive bytes) <= threshold,
                       use ranged per-member access; otherwise download the whole archive.
     :param central_directory_overhead: Passed to s3_estimate_archive_ranged_requests.
     :param member_header_overhead: Passed to s3_estimate_archive_ranged_requests.
+    :param use_chunked_reads: If ``True`` and ranged access is chosen, group adjacent members into single ranged reads.
 
     :return: An iterable of callables that return file-like objects for each requested member.
     """
+    if mode not in ("r", "rb"):
+        raise ValueError("mode must be either 'r' or 'rb'")
+    if member_header_overhead < ZIPFILE_HEADER_MIN_SIZE:
+        raise ValueError(f"member_header_overhead must be at least {ZIPFILE_HEADER_MIN_SIZE} bytes")
+
     s3_options = s3_options_from_s3_client(client)
 
-    ranged_total, archive_size, missing = s3_estimate_archive_ranged_requests(
-        client,
-        archive_url,
-        members,
-        central_directory_overhead=central_directory_overhead,
-        member_header_overhead=member_header_overhead,
-    )
+    archive_size, member_zip_infos, missed_members = s3_archive_listfile(client, bucket, key, members)
 
-    if len(missing) > 0:
-        raise FileNotFoundError(f"archive missing members: {missing}")
+    if len(missed_members) > 0:
+        raise FileNotFoundError(f"members not found in archive '{missed_members}'")
+
+    estimated_ranged_total_size = central_directory_overhead + sum(info.compress_size + member_header_overhead
+                                                                   for info in member_zip_infos)
 
     # Avoid division by zero; prefer ranged if archive size is zero (degenerate case)
-    use_ranged = True
-    if archive_size > 0:
-        use_ranged = (ranged_total / archive_size) <= threshold
+    use_ranged = (estimated_ranged_total_size / archive_size) <= threshold if archive_size > 0 else True
 
-    if use_ranged:
-        # Use fsspec zip:// style per-member access (this will issue ranged requests)
-        for member in members:
-            opener = functools.partial(s3_open_archive_member, client, archive_url, member, mode)
-            yield member, opener
+    @dataclasses.dataclass(frozen=True)
+    class MemberChunk(object):
+        name: str
+        header_offset: int
+        compress_size: int
+        compress_type: int
+
+        @property
+        def end(self) -> int:
+            return self.header_offset + self.compress_size + member_header_overhead
+
+    if use_ranged and not use_chunked_reads:
+        for info in member_zip_infos:
+            opener = functools.partial(s3_archive_open_member, client, bucket, key, info.filename, mode)
+            yield info.filename, opener
+        return
+
+    archive_url = f"s3://{bucket}/{key}"
+
+    if use_ranged and use_chunked_reads:
+        # Open archive once to read central directory and gather ZipInfo for requested members.
+        # We will group adjacent members (by local header offsets) and issue one ranged read per group,
+        # then extract each member from the group's bytes to avoid many small ranged requests.
+        with fsspec.open(archive_url, "rb", s3=s3_options) as s3_fh, zipfile.ZipFile(s3_fh) as archive:
+
+            chunks = [MemberChunk(info.filename, info.header_offset, info.compress_size, info.compress_type)
+                      for info in member_zip_infos]
+
+            chunks_groups = chunk_between(sorted(chunks, key=lambda x: x.header_offset),
+                                          chunk_func=lambda x, y: y.header_offset > x.end + member_header_overhead)
+
+            # For each group, create openers for members inside the group.
+            for group in chunks_groups:
+                group_offset = head(group).header_offset
+                group_size = last(group).end - group_offset
+
+                # Read group's bytes from remote (single ranged read)
+                s3_fh.seek(group_offset)
+                group_bytes = s3_fh.read(group_size)
+
+                def make_opener(chunk: MemberChunk) -> Callable[[], typing.IO]:
+
+                    def opener() -> typing.IO:
+                        index = chunk.header_offset - group_offset
+
+                        if index + ZIPFILE_HEADER_MIN_SIZE > len(group_bytes):
+                            raise IOError("unexpected short read of member header")
+
+                        fn_len = int.from_bytes(group_bytes[zipfile_hdr_fn_len_slice(index)], "little")
+                        ex_len = int.from_bytes(group_bytes[zipfile_hdr_ex_len_slice(index)], "little")
+
+                        raw_data_begin = index + ZIPFILE_HEADER_MIN_SIZE + fn_len + ex_len
+                        raw_data_end = raw_data_begin + chunk.compress_size
+
+                        if raw_data_end > len(group_bytes):
+                            raise IOError("unexpected short read of compressed data")
+
+                        raw_data = group_bytes[raw_data_begin:raw_data_end]
+
+                        if chunk.compress_type == zipfile.ZIP_STORED:
+                            pass
+                        elif chunk.compress_type == zipfile.ZIP_DEFLATED:
+                            raw_data = zlib.decompress(raw_data, -zlib.MAX_WBITS)
+                        else:
+                            raise NotImplementedError(f"unsupported compression '{chunk.compress_type}'")
+
+                        if mode == "r":
+                            return io.TextIOWrapper(io.BytesIO(raw_data), encoding="utf-8")
+                        else:
+                            return io.BytesIO(raw_data)
+
+                    return opener
+
+                yield from ((chunk.name, make_opener(chunk)) for chunk in group)
         return
 
     # Download full archive once and serve members from it (read member bytes into memory)
-    with fsspec.open(archive_url, "rb", s3=s3_options) as fh, tempfile.TemporaryFile() as temp_fh:
-        shutil.copyfileobj(fh, temp_fh)
+    with fsspec.open(archive_url, "rb", s3=s3_options) as s3_fh, tempfile.TemporaryFile() as temp_fh:
+        shutil.copyfileobj(s3_fh, temp_fh)
         temp_fh.seek(0)
         with zipfile.ZipFile(temp_fh) as archive:
-            for member in members:
+            for info in member_zip_infos:
                 try:
                     if mode == "r":
-                        opener = lambda m=member: io.TextIOWrapper(archive.open(m), encoding="utf-8")
+                        opener = lambda fn=info.filename: io.TextIOWrapper(archive.open(fn), encoding="utf-8")
                     else:
-                        opener = lambda m=member: io.BytesIO(archive.read(m))
-                    yield member, opener
+                        opener = lambda fn=info.filename: archive.open(fn)
+                    yield info.filename, opener
                 except KeyError as e:
                     # Shouldn't happen due to earlier check, but guard anyway
-                    raise FileNotFoundError(member) from e
+                    raise FileNotFoundError(info.filename) from e
