@@ -488,21 +488,21 @@ class S3UtilsTest(unittest.TestCase):
             local_root = os.path.join(resources_directory, "unittest", "s3utils")
             local_members = [os.path.relpath(file_path, local_root) for file_path in listfile(local_root)]
 
-            for archive_url, members in itertools.product(
-                ["s3://dummy-bucket/s3utils_archive/archive.uncompressed.zip",
-                 "s3://dummy-bucket/s3utils_archive/archive.compressed.zip"],
+            for archive_key, members in itertools.product(
+                ["s3utils_archive/archive.uncompressed.zip", "s3utils_archive/archive.compressed.zip"],
                 [local_members, None],
             ):
-                archive_size, member_zip_infos, missed_members = s3_archive_listfile(client, archive_url, members)
+                archive_size, member_zip_infos, missed_members = s3_archive_listfile(client,
+                                                                                     "dummy-bucket",
+                                                                                     archive_key,
+                                                                                     members)
 
                 self.assertEqual(archive_size - sum(info.compress_size for info in member_zip_infos), 2470)
                 self.assertEqual(set(info.filename for info in member_zip_infos), set(local_members))
                 self.assertEqual(len(missed_members), 0)
 
-            for archive_url, (members, members_expect) in itertools.product(
-                ["s3://dummy-bucket/s3utils_archive/archive.uncompressed.zip",
-                 "s3://dummy-bucket/s3utils_archive/archive.compressed.zip",
-                 ],
+            for archive_key, (members, members_expect) in itertools.product(
+                ["s3utils_archive/archive.uncompressed.zip", "s3utils_archive/archive.compressed.zip"],
                 [(local_members, local_members),
                  (None, local_members),
                  (["dir.baz/", "dir.foo/"],
@@ -524,7 +524,10 @@ class S3UtilsTest(unittest.TestCase):
                   ),
                  ],
             ):
-                archive_size, member_zip_infos, missed_members = s3_archive_listfile(client, archive_url, members)
+                archive_size, member_zip_infos, missed_members = s3_archive_listfile(client,
+                                                                                     "dummy-bucket",
+                                                                                     archive_key,
+                                                                                     members)
 
                 self.assertEqual(sorted(list(info.filename for info in member_zip_infos)), sorted(members_expect))
                 self.assertEqual(len(missed_members), 0)
@@ -545,16 +548,14 @@ class S3UtilsTest(unittest.TestCase):
             local_root = os.path.join(resources_directory, "unittest", "s3utils")
             local_members = [os.path.relpath(file_path, local_root) for file_path in listfile(local_root)]
 
-            for archive_url, mode in itertools.product(
-                ["s3://dummy-bucket/s3utils_archive/archive.uncompressed.zip",
-                 "s3://dummy-bucket/s3utils_archive/archive.compressed.zip",
-                 ],
+            for archive_key, mode in itertools.product(
+                ["s3utils_archive/archive.uncompressed.zip", "s3utils_archive/archive.compressed.zip"],
                 ["r", "rb"],
             ):
                 for local_member in local_members:
                     with (
                         open(os.path.join(local_root, local_member), mode) as local_fh,
-                        s3_archive_open_member(client, archive_url, local_member, mode) as s3_fh,
+                        s3_archive_open_member(client, "dummy-bucket", archive_key, local_member, mode) as s3_fh,
                     ):
                         self.assertEqual(local_fh.read(), s3_fh.read())
 
@@ -574,10 +575,8 @@ class S3UtilsTest(unittest.TestCase):
             local_root = os.path.join(resources_directory, "unittest", "s3utils")
             local_members = [os.path.relpath(file_path, local_root) for file_path in listfile(local_root)]
 
-            for archive_url, members, mode, threshold, use_chunked_reads in itertools.product(
-                ["s3://dummy-bucket/s3utils_archive/archive.uncompressed.zip",
-                 "s3://dummy-bucket/s3utils_archive/archive.compressed.zip",
-                 ],
+            for archive_key, members, mode, threshold, use_chunked_reads in itertools.product(
+                ["s3utils_archive/archive.uncompressed.zip", "s3utils_archive/archive.compressed.zip"],
                 [local_members, None],
                 ["r", "rb"],
                 [1.0, 0.0],
@@ -585,7 +584,8 @@ class S3UtilsTest(unittest.TestCase):
             ):
                 for member, opener in s3_archive_open_members(
                     client,
-                    archive_url,
+                    "dummy-bucket",
+                    archive_key,
                     members,
                     mode,
                     threshold=threshold,
